{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module Deus.Reactive where

import System.IO
import Data.IORef

import Data.Monoid
import Data.Maybe
import Data.Char

import Control.Applicative
import Control.Monad
import Control.Monad.Reader
import Control.Monad.Instances

import Control.Concurrent

import Data.Time.Clock
import Data.Time.Calendar
import Data.Time.Format

-- Implementation of reactive concepts and some basic combinators

-- A Future : a value known at some point in the future
  
-- simplistic IO implementation. Invariant required for this
-- implementation: blockingReadFuture will block until the future is
-- available. After the future is available, it will always return the
-- same pair (time,value).

-- the best way to implement this is with an MVar, but that wouldn't
-- be a functor, so we provide a 'smart constructor' which does
-- use the MVar

-- As useful efficiency shortcuts, we make 'Never' and 'Always'
-- constructors
  
data Future a = Never 
              | Always a
              | IOFuture (IO (UTCTime,a)) (IO (Maybe (UTCTime,a)))

dbgFutureSort Never = "Never"
dbgFutureSort (Always _) = "Always"
dbgFutureSort (IOFuture _ _) = "IOFuture"

beforeTheDawnOfTime = UTCTime (ModifiedJulianDay 0) 0

blockingReadFuture :: Future a -> IO (UTCTime,a)
blockingReadFuture Never = hangIO
blockingReadFuture (Always a) = return (beforeTheDawnOfTime,a)
blockingReadFuture (IOFuture brf _) = brf

nonblockingReadFuture :: Future a -> IO (Maybe (UTCTime,a))
nonblockingReadFuture Never = return Nothing
nonblockingReadFuture (Always a) = return (Just (beforeTheDawnOfTime,a))
nonblockingReadFuture (IOFuture _ nrf) = nrf

instance Functor Future where
  fmap f Never = Never
  fmap f (Always a) = Always (f a)
  fmap f (IOFuture brf nrf) = 
    IOFuture ((fmap.fmap) f brf) ((fmap.fmap.fmap) f nrf)

-- we don't plan to export constructors, but we do want to export these
noFuture = Never
always = Always

withTimeF :: Future a -> Future (UTCTime,a) 
withTimeF Never = Never
withTimeF (Always a) = Always (beforeTheDawnOfTime,a)
withTimeF (IOFuture brf nrf) = 
  IOFuture (fmap dblup brf) ((fmap.fmap) dblup nrf)
  where dblup (t,a) = (t,(t,a))
        
joinF :: Future (Future a) -> Future a
joinF Never = Never
joinF (Always f) = f
joinF f@(IOFuture _ _)  = 
  IOFuture 
  (do
      (t,g) <- blockingReadFuture f
      (t',a) <- blockingReadFuture g
      return (t `max` t',a))
  (do
      mbf <- nonblockingReadFuture f
      case mbf of 
        Nothing -> return Nothing
        Just (t,g) -> do          
          mbg <- nonblockingReadFuture g
          case mbg of
            Nothing -> return Nothing
            Just (t',a) -> return $ Just (t `max` t',a))

-- Monoid instance. f `mappend` g is the earlier of f and g. 
-- Left biased in the case of simultaneous futures

instance Monoid (Future a) where
  mempty = Never
  mappend Never g = g
  mappend f Never = f
  mappend (Always a) _ = Always a
  mappend _ (Always a) = Always a
  mappend f g = 
    IOFuture 
    (do
        v <- newEmptyMVar
        forkIO (blockingReadFuture f >> putMVar v ())
        forkIO (blockingReadFuture g >> putMVar v ())
        takeMVar v 
        (fromMaybe (error $ "mappend (Future a) : this can't happen " ++
                    "when mappending " ++ dbgFutureSort f ++ " with " ++
                    dbgFutureSort g
                   ) )
          `fmap` nonblockingReadFuture (mappend f g))
    -- thanks to an observation from in IRC, this implementation of the
    -- non-blocking read is not actually quite safe. It is possible for
    -- f to become ready before g, but after the call to 
    -- nonblockingReadFuture f. I think STM is the answer
    (do
        a <- nonblockingReadFuture f
        b <- nonblockingReadFuture g
        return $ case (a,b) of 
          (Just (ft,fa),Nothing) -> Just (ft,fa)
          (Nothing,Just (gt,ga)) -> Just (gt,ga)
          (Just (ft,fa),Just (gt,ga)) ->
            case compare ft gt of
              LT -> Just (ft,fa)
              GT -> Just (gt,ga)
              EQ -> Just (ft,fa)
          (Nothing,Nothing) -> Nothing)
  
-- build a future which will evaluate the given IO action 
-- asynchronously and only once

-- buildFuture's asynchronous nature -- using forkIO -- makes it ideal
-- for asynchronous/external events like socket reads, but
-- inappropriate for internal/synchronous events, as the nature of
-- thread scheduling means the future may be slightly delayed
  
buildFuture :: IO a -> IO (Future a)
buildFuture act = do
  mv <- newEmptyMVar
  sv <- newMVar False
  forkIO $ do v <- act
              t <- getCurrentTime
              putMVar mv (t,v)
              swapMVar sv True
              return ()
  return (IOFuture (readMVar mv) (do s <- readMVar sv
                                     if s then 
                                       Just `fmap` readMVar mv
                                       else
                                       return Nothing))
  
-- My One-MVar implementation didn't work, tryReadMVar (below) races
--  with itself and can return false negatives
-- tryReadMVar mv = do
--   m <- tryTakeMVar mv
--   case m of Just v -> do 
--               putMVar mv v
--               return (Just v)
--             Nothing -> return Nothing
            
-- Implement events and reactives using Futures, following Elliott
  
newtype Event a = Event (Future (Reactive a))
data Reactive a = a `Stepper` Event a

instance Functor Event where
  fmap f (Event ft) = Event ((fmap.fmap) f ft)
  
instance Functor Reactive where
  fmap f (v `Stepper` e) = f v `Stepper` fmap f e
  
-- these 'lowlevel' adapters is a way to interface events with IO
blockingGetOccurrence :: Event a -> IO (UTCTime,a,Event a)
blockingGetOccurrence (Event ft) = do
  (t,v `Stepper` e) <- blockingReadFuture ft
  return (t,v,e)

nonblockingGetOccurrence :: Event a -> IO (Maybe (UTCTime,a,Event a))
nonblockingGetOccurrence (Event ft) =
  (fmap.fmap) (\(t,v `Stepper` e) -> (t,v,e)) (nonblockingReadFuture ft)
                                     
nonblockingGetAvailableOccurrences :: Event a -> IO ([(UTCTime,a)],Event a)
nonblockingGetAvailableOccurrences ev = do
  mb <- nonblockingGetOccurrence ev
  case mb of Nothing -> return ([],ev)
             Just (t,v,fv) -> do 
               (ll,gv) <- nonblockingGetAvailableOccurrences fv
               return ((t,v):ll,gv)
               
-- similarly we need a way to interface reactives to IO
sampleR :: Reactive a -> IO (a,Reactive a)
sampleR (v `Stepper` e) = do
  nbo <- nonblockingGetOccurrence e
  case nbo of
    Just (_,v',e') -> sampleR (v' `Stepper` e')
    Nothing -> return (v,v `Stepper` e)
  
-- sample within a ref, updating it
sampleRRef :: IORef (Reactive a) -> IO a
sampleRRef ref = do
  r <- readIORef ref
  (v,r') <- sampleR r
  writeIORef ref r'
  return v
  
-- runReactor runs an Event (IO ()) in the IO monad
runReactor :: Event (IO ()) -> IO ()  
runReactor e = do
  (when,act,f) <- blockingGetOccurrence e
  act
  runReactor f

-- run as much as you can without blocking
runReactorRef :: IORef (Event (IO ())) -> IO ()
runReactorRef eref = do
  e <- readIORef eref
  (ll,f) <- nonblockingGetAvailableOccurrences e
  sequence_ (map snd ll)
  writeIORef eref f
  

instance Monoid (Event a) where
  mempty = Event noFuture
  mappend e@(Event efut) f@(Event ffut) =
    Event ((fmap ef efut) `mappend` (fmap ff ffut))
    where ef (a `Stepper` e') = (a `Stepper` (e' `mappend` f))
          ff (a `Stepper` f') = (a `Stepper` (e `mappend` f'))
          

-- constantly the same value Reactive
-- just a synonym for pure, and is it even a better name?
constReactive a = a `Stepper` mempty

-- an event consisting of a single, always available, occurrence
alwaysE :: a -> Event a
alwaysE a = Event (always (constReactive a))

-- "incrementally acummulating" events and reactives
-- like iterate, but event-driven

accumR :: a -> Event (a -> a) -> Reactive a
accumR init ev = init `seq` (init `Stepper` accumE init ev)

accumE :: a -> Event (a -> a) -> Event a
accumE init (Event ft) = 
  Event (fmap (\(f `Stepper` e') -> accumR (f init) e') ft)

-- disjoint sum of events
-- note that when they occur simultaneous, you get the
-- "Left a" before the "Right b"
sumE :: Event a -> Event b -> Event (Either a b)
sumE e f = (fmap Left e) `mappend` (fmap Right f)


instance Applicative Reactive where
  pure a = a `Stepper` mempty
  (f `Stepper` fev) <*> (a `Stepper` aev) =
    fmap (\(_,_,v) -> v) $ 
    accumR (f,a,f a) (fmap (\x -> \(f,a,v) -> 
                             case x of Left f' -> (f',a,f' a)
                                       Right a' -> (f,a',f a')) 
                      (fev `sumE` aev))


withTimeE :: Event a -> Event (UTCTime, a)
withTimeE (Event f) = 
  Event(fmap (\(t,a `Stepper` e) -> (t,a) `Stepper` withTimeE e) (withTimeF f))

-- join for events, but does this make a sensible monad instance?

joinE :: Event (Event a) -> Event a
joinE (Event ft) = 
  Event . joinF $ 
  fmap (\(e `Stepper` ee) -> 
         let (Event ftft) = (e `mappend` joinE ee) in ftft)
  ft
  
-- only pick up the 'Justs' from an Event
catMaybesE :: Event (Maybe a) -> Event a
catMaybesE = joinE . fmap (maybe mempty alwaysE)

filterE :: (a -> Bool) -> Event a -> Event a
filterE p =  joinE . fmap (\x -> if p x then alwaysE x else mempty)

-- join for reactives (not the same as join for events!)

joinR :: Reactive (Reactive a) -> Reactive a
joinR ((v0 `Stepper` Event f0) `Stepper` Event ff0) =
  v0 `Stepper` Event (((\r1 -> joinR (r1 `Stepper` Event ff0)) <$> f0)
                      `mappend`
                      (joinR <$> ff0))

-- Find the actual changes in a Reactive (given Eq)

changesR :: (Eq a) => Reactive a -> Event a
changesR (a `Stepper` ev) =
  catMaybesE . fmap snd $ 
  accumE (a,Nothing) (fmap (\a -> \ (b,_) -> 
                             if (a==b) then (b,Nothing) else (a,Just a)) 
                      ev)


switcherR :: Reactive a -> Event (Reactive a) -> Reactive a
switcherR r er = joinR (r `Stepper` er)

snapshotWithRF :: (a -> b -> c) -> Reactive a -> Future b -> Future c
snapshotWithRF fn (rcv `Stepper` Event rcf) ft = 
  (fn rcv <$> ft)
  `mappend`
  joinF ((\r -> snapshotWithRF fn r ft) <$> rcf)

switcherAuxE :: Reactive a -> Event (a -> Reactive a) -> Event (Reactive a)
switcherAuxE ra (Event fb) = 
  Event $ snapshotWithRF (\v0 (vb `Stepper` eb) -> vb v0 `switcherAuxR` eb)
  ra fb
  
switcherAuxR :: Reactive a -> Event (a -> Reactive a) -> Reactive (Reactive a)
switcherAuxR ra ev = ra `Stepper` (ra `switcherAuxE` ev)

switcherFoldR :: Reactive a -> Event (a -> Reactive a) -> Reactive a
switcherFoldR ra ev = joinR (ra `switcherAuxR` ev)

-- checkpoint a Reactive at Event occurences

snapshotWithR :: (a -> b -> c) -> Reactive a -> Event b -> Event c
snapshotWithR f rc@(rcv `Stepper` rcev) ev = 
  -- using accumE build something which 'remembers' the reactive
  -- in its first component and uses the second component to store
  -- f a b, every time b occurs
  catMaybesE . fmap snd $ 
  accumE (rcv,Nothing)
  (fmap (\x -> case x of Right a -> \(_,_) -> a `seq` (a,Nothing)
                         Left b -> \(a,_) -> let g = f a b in g `seq` (a,Just g))
   (sumE ev rcev))
--   Event . MkFuture $ do
--     -- strategy :
--     -- main thread will wait for the next occurrence of 'ev'
--     -- meanwhile we monitor the 'current value' of 'rc'
--     rcmv <- newMVar rc
--     forkIO (rcMonitor rcmv rc)
--     (t,evv,ev') <- blockingGetOccurrence ev
--     rc'@(rcv `Stepper` _) <- takeMVar rcmv
--     return (t,f rcv evv `Stepper` (snapshotWithR f rc' ev'))
--   where rcMonitor mv (rcv `Stepper` rcev) = do
--           (t,v,f) <- blockingGetOccurrence rcev
--           takeMVar mv
--           putMVar mv (v `Stepper` f)
--           rcMonitor mv (v `Stepper` f)

-- cojoin provides aged versions

cojoinE :: Event a -> Event (Event a)
cojoinE e@(Event ft) = 
  Event $ (\(a `Stepper` e') -> e `Stepper` cojoinE e') <$> ft

cojoinR :: Reactive a -> Reactive (Reactive a)
cojoinR r@(s `Stepper` (Event ft)) = r `Stepper` Event (cojoinR <$> ft)

-- a combinator to gain access to all previous events
history :: Event a -> Event [a]
history = accumE [] . fmap (:)

-- a combinator to gain access to previous events
-- the first occurrence of 'lastOne e' is simultaneous
-- with the *second* occurrence of e, but contains the value of the first

lastOne :: Event a -> Event (Maybe a)
lastOne = fmap (listToMaybe . take 1 . drop 1) . history
  

-- sometimes you want more than just access to the previous value of 
-- the event. sometimes you want time stamps

--accumEwithTime :: UTCTime -> a -> Event (UTCTime -> a -> a

-- Definition of Behaviour using FunTime again following Elliott

newtype FunTime a = FunTime (UTCTime -> a) deriving (Functor,Applicative)

newtype O f g a = O (f (g a))

instance (Functor f, Functor g) => Functor (O f g) 
         where fmap f (O h) = O ( ( (fmap . fmap) f ) h )
               
instance (Applicative f, Applicative g) =>
         Applicative (O f g) where
           pure = O . pure . pure
           (O ff) <*> (O vv) = O (liftA2 (<*>) ff vv)
             
newtype Behaviour a = Behaviour (O Reactive FunTime a) deriving (Functor,Applicative)

timeFunction :: (UTCTime -> a) -> Behaviour a
timeFunction = Behaviour . O . constReactive . FunTime

currentTime :: Behaviour UTCTime
currentTime = timeFunction id

unBehaviour (Behaviour (O r)) = r

cojoinB :: Behaviour a -> Behaviour (Behaviour a)
cojoinB = Behaviour . O . fmap (pure . Behaviour . O) . cojoinR . unBehaviour

switcher :: Behaviour a -> Event (Behaviour a) -> Behaviour a
switcher b eb = 
  Behaviour (O (joinR (unBehaviour b `Stepper` (unBehaviour <$> eb))))

switcherFold :: Behaviour a -> Event (a -> Behaviour a) -> Behaviour a
switcherFold b eb = 
--  let c = b `switcher` snapshotWith (\a f -> f a) c eb in c
  Behaviour . O $ 
  (unBehaviour b) 
  `switcherFoldR` 
  (fmap (\(t,f) -> \(FunTime ft) -> unBehaviour (f(ft t)) ) . withTimeE $ eb)
  
rToB :: Reactive a -> Behaviour a
rToB = Behaviour . O . fmap (FunTime . const)

sampleB :: Behaviour a -> IO (a,Behaviour a)
sampleB (Behaviour (O r)) = do
  t <- getCurrentTime
  (FunTime f,r') <- sampleR r
  return (f t, Behaviour (O r'))
  
sampleBRef :: IORef (Behaviour a) -> IO a
sampleBRef bref = do
  b <- readIORef bref
  (v,b') <- sampleB b
  writeIORef bref b'
  return v
  

-- snapshotWith is defined using withTimeE although it could
-- just as easily be the other way around. Which is more fundamental?
snapshotWith :: (a -> b -> c) -> Behaviour a -> Event b -> Event c
snapshotWith f (Behaviour (O r)) ev = 
  snapshotWithR (\(FunTime ft) (t,a) -> f (ft t) a) r (withTimeE ev)

snapshotWith2 :: (a -> b -> c -> d) ->
                 Behaviour a -> Behaviour b -> Event c  ->
                 Event d
snapshotWith2 f ba bb ec =
  snapshotWith (uncurry f) (liftA2 (,) ba bb) ec

--strictcomma a b = a `seq` b `seq` (a,b)

-- instance Applicative Behaviour where
--   pure = Behaviour . O . pure . pure 
--   f <*> bhv = 
    
hangIO :: IO a
hangIO = threadDelay (maxBound) >> hangIO


-- keepCalling is a way to turn an oldfashioned,
-- blocking IO call which you want to call over and over again
-- into an Event. It is best used for asynchronous 'external'
-- event sources due to its use of buildFuture. 
keepCalling :: IO a -> IO (Event a)
keepCalling act = do
  ft <- buildFuture (do c <- act
                        ev <- keepCalling act
                        return (c `Stepper` ev))  
  return (Event ft)

-- For internally posted events, the best API is creating a Sink/Event
-- pair, which can be implemented rather like Control.Concurrent.Chan
-- is, as a MVar-linked list In fact, I think the only reason we can't
-- use Concurrent.Chan directly is it doesn't export a 'peekChan' and
-- 'tryPeekChan'
  
-- no separate 'Nil' case; instead use empty MVar for that.
newtype MVarLinkedList a =  MVLL (MVar (a,MVarLinkedList a))

-- front and back of the stream.
-- invariant : back is always a full mvar
-- invariant : back is always pointing at an empty MVLL
-- NB we don't make front an MVar, our chans don't support
-- a 'destructive read'
data MyChan a = MyChan (MVarLinkedList a) (MVar (MVarLinkedList a))

newMyChan = do
  mv <- newEmptyMVar
  back <- newMVar (MVLL mv)
  return (MyChan (MVLL mv) back)
  
writeMyChan (MyChan front back) val = do
  mv <- newEmptyMVar
  modifyMVar_ back $ \(MVLL mv') -> do
    putMVar mv' (val,MVLL mv)
    return (MVLL mv)

-- blocking peek
peekMyChan :: MyChan a -> IO (a,MyChan a)
peekMyChan (MyChan (MVLL mv) back) = do
  (a,mvll) <- readMVar mv
  return (a,MyChan mvll back)

-- noblocking peek
tryPeekMyChan :: MyChan a -> IO (Maybe (a,MyChan a))
tryPeekMyChan (MyChan (MVLL mv) back) = do
  -- note that this isn't really safe, it races with another copy of itself
  mb <- tryTakeMVar mv
  case mb of Nothing -> return Nothing
             Just (a,mvll) -> do
               putMVar mv (a,mvll)
               return (Just (a,MyChan mvll back))


data Sink a = Sink { sendEvent :: a -> IO () }

myChanToEvent :: MyChan (UTCTime,a) -> Event a
myChanToEvent c =
  Event
  (IOFuture
   (do ((t,a),d) <- peekMyChan c
       return (t,a `Stepper` myChanToEvent d))
   (do mb <- tryPeekMyChan c
       case mb of 
         Nothing -> return Nothing
         Just ((t,a),d) ->
           return (Just (t,a `Stepper` myChanToEvent d)))
   )
       
makeSinkEvent :: IO (Sink a, Event a) 
makeSinkEvent = do
  c <- newMyChan
  return (Sink (\a -> do t <- getCurrentTime
                         writeMyChan c (t,a)),
          myChanToEvent c)

                        
                        
getKeypresses :: IO (Event Char)
getKeypresses = keepCalling getChar
                
getTicker :: Int -> IO (Event ())
getTicker d = keepCalling (threadDelay d)

timewiseIterate :: Int -> (a -> a) -> a -> IO (Event a)
timewiseIterate d f v = do
  ft <- buildFuture (do threadDelay d
                        ev <- timewiseIterate d f (f v)
                        return (v `Stepper` ev))
  return (Event ft)

timewiseIterateR :: Int -> (a -> a) -> a -> IO (Reactive a)
timewiseIterateR d f v = (v `Stepper`) <$> timewiseIterate d f v

-- given an Event, the time since its last occurrence
-- rather arbitrarily set before the first occurrence
sinceLast :: Event a -> Behaviour NominalDiffTime
sinceLast e = 
   diffUTCTime <$> 
   currentTime <*> 
   rToB (beforeTheDawnOfTime `Stepper` (fst <$> withTimeE e))
   
-- useful test function for playing

kpTest :: (Event Char -> Event (IO ())) -> IO ()
kpTest ev = do
  hSetBuffering stdin NoBuffering
  kp <- getKeypresses
  runReactor (ev kp)