module Deus.Stars.Model where

import qualified Data.Map as M
import Data.Map(Map)
import Data.List
import Data.Ord

import Control.Applicative

-- just for the types
import Graphics.Rendering.OpenGL

import Deus.Random

import System.Random

-- pure data model of the game
-- no display or UI code

-- convenience operators for velocities and speeds
(a,b) <+> (c,d) = (a+c,b+d)
(a,b) <-> (c,d) = (a-c,b-d)
magn (a,b) = sqrt(a*a + b*b)
(a,b) `withMag` m = let sc = m/magn (a,b) in (a*sc,b*sc)
(a,b) `capMag` m = let mg = magn (a,b)
                       sc = m/mg
                   in if sc < 1 then (a*sc,b*sc) else (a,b)

data Resource = Food
              | Ore
              | Metal
              | Medecine
              | Machinery
              | Oil
              | WarpFuel
                deriving (Show,Eq,Ord,Enum,Bounded)

basePrice Food      = 10
basePrice Ore       = 5
basePrice Metal     = 15
basePrice Medecine  = 50
basePrice Machinery = 25
basePrice Oil       = 15
basePrice WarpFuel  = 80

data Star = Star {
  starX :: !GLfloat,
  starY :: !GLfloat, 
  -- if starCol is not strict, I get a massive click-based space leak
  -- in randomColour. I do not understand this since the stars are static.
  -- I certainly don't understand how it is click related.
  starCol :: !(Color4 GLfloat),
  starName :: !String,
  starPrices :: !(Map Resource Int)
  } deriving Show
            
starPos st = (starX st,starY st)

type StarID = Int

data SFShip = SFShip { sfsState :: !SFShipState,  
                       sfsCash  :: !Int,
                       sfsCargo :: !(Map Resource Int)}
              deriving (Show)

data SFShipState = Orbiting !StarID 
                 | Journeying !StarID !StarID !Int !Int 
                   deriving (Show)

type ShipID = Int

data StarField = StarField { 
  _sfstars :: !(Map StarID Star), 
  _sfships :: !(Map ShipID SFShip),
  stardate :: !Int -- partly a hack for debugging
  } 
               deriving (Show)

mkStarField :: [Star] -> [SFShip] -> StarField
mkStarField sts shs = 
  StarField {
    _sfstars = M.fromList (zip [0..] sts),
    _sfships    = M.fromList (zip [0..] shs),
    stardate = 0
    }
  
stars :: StarField -> [Star]
stars = M.elems . _sfstars
ships = M.elems . _sfships

starsWithIDs :: StarField -> [(StarID,Star)]
starsWithIDs = M.assocs . _sfstars
shipsWithIDs = M.assocs . _sfships

getStar sf id = _sfstars sf M.! id
getShip sf id = _sfships sf M.! id

shipPos :: StarField -> SFShip -> Double -> (GLfloat,GLfloat)
shipPos sf (SFShip (Orbiting m) _ _) _ = starPos . getStar sf $ m
shipPos sf (SFShip (Journeying m n t tt) _ _) dt =
  let Star x0 y0 _ _ _ = getStar sf m 
      Star x1 y1 _ _ _ = getStar sf n
      tdx              = x1-x0
      tdy              = y1-y0
      pt               = min 1 ( (fromIntegral t + realToFrac dt) / 
                                 fromIntegral tt )
      x                = x0 + tdx * pt
      y                = y0 + tdy * pt
  in (x,y)

shipAng sf (SFShip (Orbiting _) _ _) = 0
shipAng sf (SFShip (Journeying m n t tt) _ _) =
  let Star x0 y0 _ _ _ = getStar sf m 
      Star x1 y1 _ _ _ = getStar sf n
      tdx              = x1-x0
      tdy              = y1-y0
  in atan2 tdx tdy * 180 / pi

-- algorithm for ship at a planet
-- (1) sell all cargo that is sellable locally
-- (2) decide on destination
-- (3) fill up on whatever is cheaper here than there

-- precondition: ship is orbiting
sellAll :: StarField -> StarID -> SFShip -> SFShip
sellAll sf stid sh =
  let st = getStar sf stid
      proceeds = M.intersectionWith (*) (sfsCargo sh) (starPrices st)
      total    = sum . M.elems $ proceeds
      soldres  = M.keys $ proceeds
  in sh { sfsCash = sfsCash sh + total,
          sfsCargo = foldr M.delete (sfsCargo sh) soldres }

-- precondition: ship is orbiting
chooseDestination :: RandomGen g => StarField -> SFShip -> RandomM g SFShip
chooseDestination sf sh =
  let (Orbiting stid) = sfsState sh
      st        = getStar sf stid
      maxhop    = 500
      possdests = filter (\(i,t) -> i /= stid && 
                                    magn (starPos st <-> starPos t) < maxhop) $
                  starsWithIDs sf
  in do
    (did,dst) <- randomListM possdests
    let distance = magn (starPos dst <-> starPos st)
        time     = round (distance/10)
    return $! (sh { sfsState = Journeying stid did 0 time })

-- precondition: ship is at the beginnign of a journey
stockUp :: StarField -> SFShip -> SFShip
stockUp sf sh = 
  let (Journeying stid did 0 time) = sfsState sh
      here = getStar sf stid
      dest = getStar sf did
      choices = M.assocs .
             M.filter (>0) $
             M.intersectionWith (-) (starPrices dest) (starPrices here)
      res  = fst . maximumBy (comparing snd) $ choices
      price = starPrices here M.! res
      total = 50 * price
  in case choices of 
    [] -> sh
    _ -> sh { sfsCash  = sfsCash sh - total,
              sfsCargo = M.insertWith' (+) res 50 $ sfsCargo sh }
      

evolveStarField :: RandomGen g => StarField -> RandomM g StarField
evolveStarField sf = do
  ships' <- mapM (\sh -> 
                   case sfsState sh of 
                     Orbiting n -> stockUp sf <$> chooseDestination sf sh
                     Journeying n m t tt ->
                       return 
                       (if t + 1 > tt then
                          sellAll sf m (sh { sfsState = Orbiting m} )
                        else
                          sh { sfsState = Journeying n m (t+1) tt })
                 ) (ships sf)
  (foldr seq () ships') `seq` return sf {_sfships = M.fromList (zip [0..] ships'),stardate = stardate sf + 1}
